// Copyright © 2023 ab180. All rights reserved.
#include "AirbridgeAndroidCallback.h"

#if PLATFORM_ANDROID
#include "AirbridgeAndroidCore.h"
#include "AirbridgeUnrealModule.h"
#include "Platforms/Android/Util/AirbridgeJNI.h"

#define AIRBRIDGE_UNREAL_MODULE_POINTER		(FAirbridgeUnrealModule::Get())
#define ANDROID_CORE_POINTER				(static_cast<AirbridgeAndroidCore*>((AIRBRIDGE_UNREAL_MODULE_POINTER)->GetAirbridge().Get()))

jint GetCallbackType(JNIEnv *Env, jobject This)
{
	jclass airbridgeCallbackClass = Env->GetObjectClass(This);
	jmethodID getCallbackTypeMethodId = Env->GetMethodID(airbridgeCallbackClass, "getCallbackType", "()I");
	if (getCallbackTypeMethodId == nullptr) {
		return -1;
	}
	jint type = Env->CallIntMethod(This, getCallbackTypeMethodId);
	return type;
}

FString GetCallbackUUID(JNIEnv *Env, jobject This)
{
	jclass airbridgeCallbackClass = Env->GetObjectClass(This);
	jmethodID getCallbackUUIDMethodId = Env->GetMethodID(airbridgeCallbackClass, "getUUID", "()Ljava/lang/String;");
	if (getCallbackUUIDMethodId == nullptr) {
		return "";
	}
	jstring uuidString = (jstring) Env->CallObjectMethod(This, getCallbackUUIDMethodId);
	return ConvertString(Env, uuidString);
}

bool IsBroadcastCallback(JNIEnv *Env, jobject This)
{
	jclass airbridgeCallbackClass = Env->GetObjectClass(This);
	jmethodID isBroadcastMethodId = Env->GetMethodID(airbridgeCallbackClass, "isBroadcast", "()Z");
	if (isBroadcastMethodId == nullptr) {
		return false;
	}
	jboolean result = Env->CallBooleanMethod(This, isBroadcastMethodId);
	return (result != JNI_FALSE);
}

FString ConvertString(JNIEnv *Env, jstring String)
{
	// Copies the new key into its final C string buffer
	const char* cUriString = Env->GetStringUTFChars(String, nullptr);
	if (cUriString == nullptr) {
		return "";  /* OutOfMemoryError already thrown */
	}
	
	FString toReturn = FString(cUriString);

	//release heap memory
	Env->ReleaseStringUTFChars(String, cUriString);
	Env->DeleteLocalRef(String);

	return toReturn;
}

FString GetUriString(JNIEnv *Env, jobject Result)
{
	jclass uriClass = Env->GetObjectClass(Result);
	jmethodID toStringMethodId = Env->GetMethodID(uriClass, "toString", "()Ljava/lang/String;");
	jstring uriString = (jstring) Env->CallObjectMethod(Result, toStringMethodId);
	
	return ConvertString(Env, uriString);
}

extern "C" {
	/*
	 * Class:     co_ab180_airbridge_ue_callback_AirbridgeOnSuccessCallback
	 * Method:    invoke
	 * Signature: (Ljava/lang/Object;)V
	 */
	JNIEXPORT void JNICALL Java_co_ab180_airbridge_ue_callback_AirbridgeOnSuccessCallback_invoke
	(JNIEnv *Env, jobject This, jobject Result)
	{
		FString UUID = GetCallbackUUID(Env, This);				// Get Method UUID
		bool IsBroadcast = IsBroadcastCallback(Env, This);		// Is Broadcast
		
		switch (GetCallbackType(Env, This)) {
		case HandleDeeplink:
			// Check if module is loaded
			if (AIRBRIDGE_UNREAL_MODULE_POINTER)
			{
				ANDROID_CORE_POINTER->ProcessHandleDeeplink(GetUriString(Env, Result));
			}
			// If module is not loaded
			else
			{
				AirbridgeAndroidCore::StartDeeplink = GetUriString(Env, Result);
			}
			break;
		case HandleDeferredDeeplink:
			if (Result == nullptr) return;	// nullable
			
			if (AIRBRIDGE_UNREAL_MODULE_POINTER)
				ANDROID_CORE_POINTER->ProcessHandleDeferredDeeplink(GetUriString(Env, Result));	
			break;
		case Click:
			if (AIRBRIDGE_UNREAL_MODULE_POINTER)
				ANDROID_CORE_POINTER->ProcessClickOnSuccess(UUID);
			break;
		case Impression:
			if (AIRBRIDGE_UNREAL_MODULE_POINTER)
				ANDROID_CORE_POINTER->ProcessImpressionOnSuccess(UUID);
			break;
		case FetchDeviceUUID:
			if (AIRBRIDGE_UNREAL_MODULE_POINTER)
			{
				if (IsBroadcast)
				{
					ANDROID_CORE_POINTER->UnrealBroadcastDeviceUUID(ConvertString(Env, (jstring)Result));
				}
				else
				{
					ANDROID_CORE_POINTER->ProcessFetchDeviceUUIDOnSuccess(UUID, ConvertString(Env, (jstring)Result));
				}
			}
			break;
		case FetchAirbridgeGeneratedUUID:
			if (AIRBRIDGE_UNREAL_MODULE_POINTER)
			{
				if (IsBroadcast)
				{
					ANDROID_CORE_POINTER->UnrealBroadcastAirbridgeGeneratedUUID(ConvertString(Env, (jstring)Result));
				}
				else
				{
					ANDROID_CORE_POINTER->ProcessFetchAirbridgeGeneratedUUIDOnSuccess(UUID, ConvertString(Env, (jstring)Result));
				}
			}
			break;
		default:
			break;
		}
	}

	/*
	 * Class:     co_ab180_airbridge_ue_callback_AirbridgeOnFailureCallback
	 * Method:    invoke
	 * Signature: (Ljava/lang/Throwable;)V
	 */
	JNIEXPORT void JNICALL Java_co_ab180_airbridge_ue_callback_AirbridgeOnFailureCallback_invoke
	(JNIEnv *Env, jobject This, jthrowable Result)
	{
		// Find the Throwable class (java/lang/Throwable)
		jclass throwableClass = Env->GetObjectClass(Result);
		if (!throwableClass) return;
		
		// Get the getMessage method ID
		jmethodID getMessageMethodId = Env->GetMethodID(throwableClass, "getMessage", "()Ljava/lang/String;");
		if (!getMessageMethodId) return;
		
		// Call the getMessage method on the throwable
		jstring jMessage = (jstring) Env->CallObjectMethod(Result, getMessageMethodId);
		if (!jMessage) return;
		FString Message = ConvertString(Env, jMessage);

		// Get Method UUID
		FString UUID = GetCallbackUUID(Env, This);
		
		switch (GetCallbackType(Env, This)) {
		case HandleDeeplink:
			break;
		case HandleDeferredDeeplink:
			break;
		case Click:
			if (AIRBRIDGE_UNREAL_MODULE_POINTER)
				ANDROID_CORE_POINTER->ProcessClickOnFailure(UUID, Message);
			break;
		case Impression:
			if (AIRBRIDGE_UNREAL_MODULE_POINTER)
				ANDROID_CORE_POINTER->ProcessImpressionOnFailure(UUID, Message);
			break;
		case FetchDeviceUUID:
			if (AIRBRIDGE_UNREAL_MODULE_POINTER)
				ANDROID_CORE_POINTER->ProcessFetchDeviceUUIDOnFailure(UUID, Message);
			break;
		case FetchAirbridgeGeneratedUUID:
			if (AIRBRIDGE_UNREAL_MODULE_POINTER)
				ANDROID_CORE_POINTER->ProcessFetchAirbridgeGeneratedUUIDOnFailure(UUID, Message);
			break;
		default:
			break;
		}
	}

	/*
	 * Class:     co_ab180_airbridge_ue_callback_AirbridgeOnAttributionResultReceiveListener
	 * Method:    onAttributionResultReceived
	 * Signature: (Ljava/util/Map;)V
	 */
	JNIEXPORT void JNICALL Java_co_ab180_airbridge_ue_callback_AirbridgeOnAttributionResultReceiveListener_onAttributionResultReceived
	(JNIEnv *Env, jobject This, jobject AttributionResult)
	{
		TMap<FString, FString> AttributionResultData;
		if (AirbridgeJNI::JavaMapToUnrealTMap(Env, AttributionResult, AttributionResultData))
		{
			// Check if module is loaded
			if (AIRBRIDGE_UNREAL_MODULE_POINTER)
			{
				ANDROID_CORE_POINTER->ProcessOnAttributionResultReceived(AttributionResultData);
			}
			// If module is not loaded
			else
			{
				AirbridgeAndroidCore::ReceivedAttribution = AttributionResultData;
			}
		}
	}
	
	/*
	 * Class:     co_ab180_airbridge_ue_callback_AirbridgeHandleDeeplink
	 * Method:    handleDeeplink
	 * Signature: (Ljava/lang/String;)V
	 */
	JNIEXPORT void JNICALL Java_co_ab180_airbridge_ue_callback_AirbridgeHandleDeeplink_handleDeeplink
	(JNIEnv *Env, jobject This, jstring Deeplink)
	{
		if (Deeplink == nullptr) return;

		// Check if module is loaded
		if (AIRBRIDGE_UNREAL_MODULE_POINTER)
		{
			ANDROID_CORE_POINTER->ProcessHandleDeeplink(GetUriString(Env, Deeplink));
		}
		// If module is not loaded
		else
		{
			AirbridgeAndroidCore::StartDeeplink = GetUriString(Env, Deeplink);
		}
	}
}
#endif